# Copyright (C) 2022 Dremio Corporation

# Copyright (c) 2019 Ryan Murray.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import requests

from dbt.adapters.dremio.api.authentication import DremioPatAuthentication
from dbt.adapters.dremio.api.parameters import Parameters
from dbt.adapters.dremio.api.rest.utils import _post, _get, _put, _delete
from dbt.adapters.dremio.api.rest.url_builder import UrlBuilder

from dbt.adapters.events.logging import AdapterLogger

logger = AdapterLogger("dremio")

session = requests.Session()


class DremioRestClient:
    def __init__(self, api_parameters: Parameters):
        self._parameters = api_parameters

    def start(self):
        self._parameters = self.__login()

    def __login(self, timeout=10):
        if isinstance(self._parameters.authentication, DremioPatAuthentication):
            return self._parameters

        url = UrlBuilder.login_url(self._parameters)
        response = _post(
            url,
            json={
                "userName": self._parameters.authentication.username,
                "password": self._parameters.authentication.password,
            },
            timeout=timeout,
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

        self._parameters.authentication.token = response["token"]

        return self._parameters

    def sql_endpoint(self, query, context=None):
        url = UrlBuilder.sql_url(self._parameters)
        return _post(
            url,
            self._parameters.authentication.get_headers(),
            ssl_verify=self._parameters.authentication.verify_ssl,
            json={"sql": query, "context": context},
        )

    def job_status(self, job_id):
        url = UrlBuilder.job_status_url(self._parameters, job_id)
        return _get(
            url,
            self._parameters.authentication.get_headers(),
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def job_cancel_api(self, job_id):
        url = UrlBuilder.job_cancel_url(self._parameters, job_id)
        return _post(
            url,
            self._parameters.authentication.get_headers(),
            json=None,
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def job_results(self, job_id, offset=0, limit=100):
        url = UrlBuilder.job_results_url(
            self._parameters,
            job_id,
            offset,
            limit,
        )
        return _get(
            url,
            self._parameters.authentication.get_headers(),
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def create_catalog_api(self, json):
        url = UrlBuilder.catalog_url(self._parameters)
        return _post(
            url,
            self._parameters.authentication.get_headers(),
            json=json,
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def get_catalog_item(self, catalog_id=None, catalog_path=None):
        if catalog_id is None and catalog_path is None:
            raise TypeError(
                "both id and path can't be None for a catalog_item call")

        # Will use path if both id and path are specified
        if catalog_path:
            url = UrlBuilder.catalog_item_by_path_url(
                self._parameters,
                catalog_path,
            )
        else:
            url = UrlBuilder.catalog_item_by_id_url(
                self._parameters,
                catalog_id,
            )
        return _get(
            url,
            self._parameters.authentication.get_headers(),
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def delete_catalog(self, cid):
        url = UrlBuilder.delete_catalog_url(self._parameters, cid)
        return _delete(
            url,
            self._parameters.authentication.get_headers(),
            ssl_verify=self._parameters.authentication.verify_ssl,
        )
    
    # dbt docs integration within Dremio wikis and tags
    def create_wiki(self, object_id: str, text: str):
        url = UrlBuilder.wikis_management_url(self._parameters, object_id)
        return _post(
            url,
            self._parameters.authentication.get_headers(),
            json={"text": text},
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def retrieve_wiki(self, object_id: str):
        url = UrlBuilder.wikis_management_url(self._parameters, object_id)
        return _get(
            url,
            self._parameters.authentication.get_headers(),
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def update_wiki(self, object_id: str, text: str, version: int):
        url = UrlBuilder.wikis_management_url(self._parameters, object_id)
        return _post(
            url,
            self._parameters.authentication.get_headers(),
            json={"text": text, "version": version},
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def delete_wiki(self, object_id: str, version: int):
        url = UrlBuilder.wikis_management_url(self._parameters, object_id)
        return _post(
            url,
            self._parameters.authentication.get_headers(),
            json={"text": "", "version": version},
            ssl_verify=self._parameters.authentication.verify_ssl,
        )


    def create_tags(self, dataset_id: str, tags: list[str]):
        url = UrlBuilder.tags_management_url(self._parameters, dataset_id)
        return _post(
            url,
            self._parameters.authentication.get_headers(),
            json={"tags": tags},
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def retrieve_tags(self, dataset_id: str):
        url = UrlBuilder.tags_management_url(self._parameters, dataset_id)
        return _get(
            url,
            self._parameters.authentication.get_headers(),
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def update_tags(self, dataset_id: str, tags: list[str], version: str):
        url = UrlBuilder.tags_management_url(self._parameters, dataset_id)
        return _post(
            url,
            self._parameters.authentication.get_headers(),
            json={"tags": tags, "version": version},
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def delete_tags(self, dataset_id: str, version: str):
        url = UrlBuilder.tags_management_url(self._parameters, dataset_id)
        return _post(
            url,
            self._parameters.authentication.get_headers(),
            json={"tags": [], "version": version},
            ssl_verify=self._parameters.authentication.verify_ssl,
        )


    def get_reflections(self, dataset_id):
        url = UrlBuilder.get_reflection_url(self._parameters, dataset_id)
        return _get(
            url,
            self._parameters.authentication.get_headers(),
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def create_reflection(self, payload):
        url = UrlBuilder.create_reflection_url(self._parameters)
        return _post(
            url,
            self._parameters.authentication.get_headers(),
            json=payload,
            ssl_verify=self._parameters.authentication.verify_ssl,
        )

    def update_reflection(self, reflection_id, payload):
        url = UrlBuilder.update_reflection_url(self._parameters, reflection_id)
        return _put(
            url,
            self._parameters.authentication.get_headers(),
            json=payload,
            ssl_verify=self._parameters.authentication.verify_ssl,
        )
    


from enum import Enum


class TransformType(Enum):
    YEAR = "YEAR"
    MONTH = "MONTH"
    DAY = "DAY"
    HOUR = "HOUR"
    IDENTITY = "IDENTITY"
    BUCKET = "BUCKET"
    TRUNCATE = "TRUNCATE"

    @classmethod
    def from_string(cls, transform_str):
        transform_str = transform_str.upper()

        if transform_str.startswith("BUCKET("):
            return cls.BUCKET
        elif transform_str.startswith("TRUNCATE("):
            return cls.TRUNCATE

        try:
            return cls(transform_str)
        except ValueError:
            return cls.IDENTITY

    def to_transform(self, raw_str):
        if self in (
                TransformType.YEAR,
                TransformType.MONTH,
                TransformType.DAY,
                TransformType.HOUR,
                TransformType.IDENTITY
        ):
            return {"type": self.value}

        if self == TransformType.BUCKET:
            bucket_count = int(raw_str.split("(")[1].split(")")[0])
            return {
                "type": "BUCKET",
                "bucketTransform": {"bucketCount": bucket_count},
            }

        if self == TransformType.TRUNCATE:
            truncate_length = int(raw_str.split("(")[1].split(")")[0])
            return {
                "type": "TRUNCATE",
                "truncateTransform": {"truncateLength": truncate_length},
            }

        return {"type": TransformType.IDENTITY.value}


# https://docs.dremio.com/current/reference/api/reflections/
class ReflectionEntity:
    def __init__(self, name, reflection_type, dataset_id, display_fields, dimensions, date_dimensions, measures,
                 computations, partition_by, partition_transform, partition_method, distribute_by, localsort_by,
                 arrow_cache):
        self.__name = name
        self.__type = reflection_type
        self.__dataset_id = dataset_id
        self.__partition_method = partition_method.upper()
        self.__display_fields = display_fields
        self.__dimensions_fields = dimensions
        self.__date_dimensions_fields = date_dimensions
        self.__measure_fields = measures
        self.__computation_fields = computations
        self.__partition_by_fields = partition_by
        self.__partition_transformations = partition_transform
        self.__partition_method = partition_method
        self.__distribute_by_fields = distribute_by
        self.__local_sort_fields = localsort_by
        self.__arrow_cache = arrow_cache

    def buildDisplayFields(self):
        return [{"name": field} for field in self.__display_fields]

    def buildDimensionFields(self):
        return [{"name": field} for field in self.__dimensions_fields]

    def buildDateFields(self):
        return [{"name": date_dimension, "granularity": "DATE"} for date_dimension in self.__date_dimensions_fields]

    def buildMeasureFields(self):
        return [{"name": measure, "measureTypeList": computation.split(',')} for
                measure, computation in zip(self.__measure_fields, self.__computation_fields)]

    def buildPartitionFields(self):
        if not self.__partition_transformations:
            self.__partition_transformations = ["IDENTITY"] * len(self.__partition_by_fields)

        partition_fields = []

        for partition, transform in zip(self.__partition_by_fields, self.__partition_transformations):
            transform_type = TransformType.from_string(transform)
            partition_fields.append({
                "name": partition,
                "transform": transform_type.to_transform(transform)
            })

        return partition_fields

    def buildDistributionFields(self):
        return [{"name": distribute} for distribute in self.__distribute_by_fields]

    def buildSortFields(self):
        return [{"name": sort} for sort in self.__local_sort_fields]

    def build_payload(self):
        payload = {
            "type": self.__type,
            "name": self.__name,
            "datasetId": self.__dataset_id,
            "enabled": True,
            "arrowCachingEnabled": self.__arrow_cache,
            "partitionDistributionStrategy": self.__partition_method.upper(),
            "entityType": "reflection"
        }

        if self.__display_fields:
            payload["displayFields"] = self.buildDisplayFields()

        if self.__dimensions_fields:
            payload["dimensionFields"] = self.buildDimensionFields()

        if self.__date_dimensions_fields:
            payload["dateFields"] = self.buildDateFields()

        if self.__measure_fields and self.__computation_fields:
            payload["measureFields"] = self.buildMeasureFields()

        if self.__partition_by_fields:
            payload["partitionFields"] = self.buildPartitionFields()

        if self.__distribute_by_fields:
            payload["distributionFields"] = self.buildDistributionFields()

        if self.__local_sort_fields:
            payload["sortFields"] = self.buildSortFields()

        return payload
